# utils.py
"""
Utils
-----
Utility functions for the whole project.
"""
import collections
from copy import deepcopy
import logging.config
import os
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.tseries import offsets

from soam.constants import DS_COL

logger = logging.getLogger(__name__)


def range_datetime(
    datetime_start,
    datetime_end,
    hourly_offset: bool = False,
    timeskip=None,
    as_datetime: bool = False,
):
    # TODO: review datetime_start, datetime_end, are datetimes?
    # TODO: timeskip is Tick?
    """
    Build datetime generator over successive time steps.

    Parameters
    ----------
    datetime_start: datetime
        Start datetime.
    datetime_end: datetime
        End datetime.
    hourly_offset: boolean
        Wheteher to offset hourly. False by default.
    timeskip: Tick
        An instance of fast-forwarding a substantial amount of time.
    as_datetime: boolean
        Whether the object type should be datetime. False by default.
    """
    if timeskip is None:
        timeskip = offsets.Day(1) if not hourly_offset else offsets.Hour(1)
    if not isinstance(datetime_start, pd.Timestamp):
        datetime_start = pd.Timestamp(datetime_start)
    while datetime_start <= datetime_end:
        if as_datetime:
            yield datetime_start.to_pydatetime()
        else:
            yield datetime_start
        datetime_start += timeskip


def sanitize_arg(v, default=None):
    """
    Sanitize mutable arguments.

    Get a sanitized version of the given argument value to avoid mutability issues.

    The default arg provided is also deepcopied to avoid problems on that part.

    To use replace this:
    ```
    def f(x={}):
        ...
    ```
    With this:
    ```
    def f(x=None):
        x = sanitize_arg(x, {})
    ```

    Parameters
    ----------
    v: Value to check.
    default: Value to set if.

    Returns
    -------
    Santized value.
    """
    if default is None:
        default = {}
    if v is None:
        return deepcopy(default)
    else:
        return v


def sanitize_arg_empty_dict(v):
    """Convenience function for `sanitize_arg(v, {})`."""
    return sanitize_arg(v, {})


def get_file_path(path: Path, fn: str) -> Path:
    """
    Find an available path for a file, using an index prefix.

    Parameters
    ----------
    path: Path
        file path
    fn: str
        filename

    Returns
    ----------
    path
        file path
    """
    paths = path.glob(f"*_{fn}")
    max_index = max((int(p.name.split("_")[0]) for p in paths), default=-1) + 1
    return path / f"{max_index}_{fn}"


def filter_by_class_or_subclass(l, c):
    """
    Filter-out objects from list that are not class or subclass of c.

    Parameters
    ----------
    l: List
        List of objects.
    c: Class
        Reference class.

    Returns
    -------
    Comprehensive list
        Looks into the given list to check if the objects inside of it are or arent objects of the reference given class or subclass of c.
        It filters the ones that arent objects of the class or subclass of c.
    """
    return [e for e in l if isinstance(e, c) or issubclass(e.__class__, c)]


def split_backtesting_ranges(
    time_series: pd.DataFrame,
    test_window: int = 1,
    train_window: Optional[int] = 1,
    step_size: int = None,
) -> List[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Generates time series partitions for backtesting.

    Parameters
    ----------
    time_series: pd.DataFrame
        Original data used for training and evaluation.
    test_window: int
        Time range steps to be extracted from the end of the original time series.
    train_window: int, optional
        Time range steps to be extracted before the test data.
        If a value is passed then the sliding method will be used to select
         the training data.
        If `None` then the full time series will be used, the expanding window method.
         It will start with the first train window of step_size size.
    step_size: int
        Distance between each successive step between the beginning of each forecasting
         range. If None defaults to test_window.

    Returns
    -------
    list of tuple of pd.DataFrame
        A list of tuples, of train_set and test_set, to be used to train or evaluate
         the model.

    Notes
    -----
    The end of the split is going to be train_window plus a multiple of step_size. So
     some of the last elements of the time series can be not used in the resulting
     splits.

    Raises
    ------
    IndexError
        If the time series is empty, if the test_window or train_window are greater than
         the time series length, or if the step size is lower than 1.

    See Also
    --------
    For Theorical background read: documentation/references###Window_policies
    """
    if step_size is None:
        step_size = test_window

    raise_message = None
    if time_series.empty:
        raise_message = "The time series dataframe is empty."
    elif 0 >= test_window or test_window >= len(time_series):
        raise_message = "Inconsistent test window."
    elif 0 >= step_size:
        raise_message = "Step size should be greater than zero."
    elif train_window is not None and (
        0 >= train_window or train_window >= len(time_series)
    ):
        raise_message = "Inconsistent train window."
    if raise_message:
        raise IndexError(raise_message)

    result_slices = []
    for start_test_pos in range(
        train_window or step_size, len(time_series) - test_window + 1, step_size
    ):
        result_slices.append(
            (
                time_series.iloc[
                    start_test_pos - train_window
                    if train_window
                    else 0 : start_test_pos
                ],
                time_series.iloc[start_test_pos : start_test_pos + test_window],
            )
        )

    return result_slices


class SuppressStdOutStdErr(object):
    """
    A context manager to perform a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.

    This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    Ref:
        https://github.com/facebook/prophet/issues/223
    """

    def __init__(self):
        """Open a pair of null files and save the actual stdout (1) and stderr (2) file descriptors."""
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)


def add_future_dates(
    df: pd.DataFrame, periods: int, frequency: str = None, ds_col: str = DS_COL,
):
    """
    Add future dates to dataframe for which to predict with.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to be appended with future dates.
    periods : int
        number of new future rows or datapoints to append.
    frequency : str, optional
        pandas frequency string, by default None
    ds_col : str, optional
        date column name, by default DS_COL

    Returns
    -------
    pandas.DataFrame
        DataFrame with original data plos new dates appended.

    Raises
    ------
    ValueError
        Frequency could not be inferred.
    """
    if frequency is None:
        inferred_frequency = pd.infer_freq(df[ds_col])
        if inferred_frequency is None:
            raise ValueError(
                "Frequency could not be inferred, please specify a frequency."
            )
        frequency = inferred_frequency
    date_range_values = pd.date_range(
        start=df[[ds_col]].iloc[-1][0], periods=periods + 1, freq=frequency
    ).values[1:]
    future_df = {col: [np.nan] * periods for col in df.columns if col != ds_col}
    future_df[ds_col] = date_range_values
    future_df = pd.DataFrame(future_df)
    full_df = pd.concat([df, future_df], axis=0, ignore_index=True)
    return full_df


def flatten_dict(d, parent_key='', sep='.'):
    """
    Flatten a nested dictionary, creating new keys for your new dictionary.

    Parameters
    ----------
    d : Dict
        dictionary to be flattened.
    parent_key : str, optional
        parent key, by default None.
    sep : str, optional
        separator, by defaut '.'.

    Returns
    -------
    Dict
        Flattened dictionary.
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



"""Test the utils."""
import os
from os import listdir
from pathlib import Path
from typing import List, Tuple
import unittest

import numpy as np
import pandas as pd
import pytest

from soam.utilities.utils import add_future_dates, split_backtesting_ranges

ROOT_TEST_DIRECTORY = Path(__file__).parent / "resources" / "test_utils"
VALIDATION_PREFIX = "validation_"
TRAIN_PREFIX = "train_"


class TestSplitBacktestingRanges(unittest.TestCase):
    """Create an object to test the utils."""

    def setUp(self):
        with open(ROOT_TEST_DIRECTORY / "initial_dataframe.csv") as df_stream:
            self.initial_df = pd.read_csv(df_stream, index_col=0)
        self.initial_df_size = len(self.initial_df)

    def test_empty_dataframe(self):
        # TODO: find how to modify default error message.
        with self.assertRaises(IndexError):
            split_backtesting_ranges(pd.DataFrame())

    def test_test_window_under_minimum_border(self):
        with self.assertRaises(IndexError):
            split_backtesting_ranges(self.initial_df, test_window=0)

    def test_test_window_over_maximum_border(self):
        with self.assertRaises(IndexError):
            split_backtesting_ranges(self.initial_df, test_window=self.initial_df_size)

    def test_train_window_under_minimum_border(self):
        with self.assertRaises(IndexError):
            split_backtesting_ranges(self.initial_df, train_window=0)

    def test_train_window_over_maximum_border(self):
        with self.assertRaises(IndexError):
            split_backtesting_ranges(self.initial_df, train_window=self.initial_df_size)

    def test_step_size_under_minimum_border(self):
        with self.assertRaises(IndexError):
            split_backtesting_ranges(self.initial_df, step_size=0)

    def test_all_default(self):
        self.template_test_against_initial_df(
            "all_default/", "default parameters failing"
        )

    def test_test_window_maximum_border(self):
        self.template_test_against_initial_df(
            "test_window_maximum_border/",
            "maximum test window value failing",
            test_window=self.initial_df_size - 1,
        )

    def test_train_window_maximum_border(self):
        self.template_test_against_initial_df(
            "train_window_maximum_border/",
            "maximum train window value failing",
            train_window=self.initial_df_size - 1,
        )

    def test_step_size_equal_to_df_len(self):
        self.template_test_against_initial_df(
            "step_size_equal_to_df_size/",
            "df len step size value failing",
            step_size=self.initial_df_size,
        )

    def test_normal_usage_sliding(self):
        self.template_test_against_initial_df(
            "test_normal_usage_sliding/",
            "normal usage with sliding window" " failing",
            test_window=2,
            train_window=2,
            step_size=2,
        )

    def test_normal_usage_expanding(self):
        self.template_test_against_initial_df(
            "test_normal_usage_expanding/",
            "normal usage with expanding window" " failing",
            test_window=2,
            train_window=None,
            step_size=2,
        )

    def template_test_against_initial_df(
        self, folder: Path, error_message: str, **split_backtesting_ranges_kwarg
    ):
        """
        Test the result of passing the initial dataframe to split_backtesting_ranges,
        using split_backtesting_ranges_kwarg, against the dataframes contained in
        folder.

        Parameters
        ----------
        folder: pathlib.Path
            Path to the directory to drawn train and validation dataframes.
        error_message: str
            Path to the directory to drawn train and validation dataframes.
        split_backtesting_ranges_kwarg: dict
            The keyword arguments for split_backtesting_ranges.

        See Also
        --------
        load_directory_dataframes : to know the file naming convention to load the train
         and validation datasets.
        """
        output = split_backtesting_ranges(
            self.initial_df, **split_backtesting_ranges_kwarg
        )
        expected_output = self.load_directory_dataframes(ROOT_TEST_DIRECTORY / folder)
        self.assertEqual(len(expected_output), len(output), error_message)
        for (
            [train_df_function, test_df_function],
            [train_df_loaded, test_df_loaded],
        ) in zip(output, expected_output):
            self.assertTrue(train_df_function.equals(train_df_loaded), error_message)
            self.assertTrue(test_df_function.equals(test_df_loaded), error_message)

    @staticmethod
    def load_directory_dataframes(
        path: Path,
    ) -> List[Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Loads splitted train-validation dataframes from a directory.

        The pair of csvs in the directory named "{TRAIN_PREFIX}{step}.csv" and
        "{VALIDATION_PREFIX}{step}.csv" would be loaded as a list of tuples of (train,
         validation) dataframes.

         TRAIN_PREFIX and VALIDATION_PREFIX constant are declared at the beginning
         of this file.

        Parameters
        ----------
        path: pathlib.Path
            Path to the directory to drawn train and validation dataframes.

        Returns
        -------
        list of tuple of pd.DataFrame
            The train and validation dataframes loaded from the directory.
        """
        # TODO: move this dataframes in csvs to code.
        directory_dataframes = dict()
        for validation_file_name in listdir(path):
            splited_file_name = os.path.splitext(validation_file_name)
            train_file_name = validation_file_name.replace(
                VALIDATION_PREFIX, TRAIN_PREFIX
            )
            if (
                (path / train_file_name).is_file()
                and len(splited_file_name) == 2
                and VALIDATION_PREFIX in splited_file_name[0]
                and splited_file_name[1] == ".csv"
            ):
                with open(path / train_file_name) as train_stream, open(
                    path / validation_file_name
                ) as validation_stream:
                    directory_dataframes[validation_file_name] = (
                        pd.read_csv(train_stream, index_col=0),
                        pd.read_csv(validation_stream, index_col=0),
                    )
        return [
            directory_dataframes[loaded_split_name]
            for loaded_split_name in sorted(directory_dataframes.keys())
        ]


def test_add_future_dates_frequency_not_inferred():
    """Test that frequency can't be inferred."""
    test_data = pd.DataFrame.from_records(
        np.array(
            [
                ('2016-06-01T00:00:00.000000000', 453990.926731),
                ('2016-07-03T00:00:00.000000000', 461908.605069),
                ('2016-08-04T00:00:00.000000000', 482625.559603),
                ('2016-09-05T00:00:00.000000000', 438345.625953),
                ('2016-10-01T00:00:00.000000000', 458133.301185),
                ('2016-11-01T00:00:00.000000000', 467090.524589),
                ('2016-12-03T00:00:00.000000000', 508706.424433),
                ('2017-12-06T00:00:00.000000000', 426140.823832),
                ('2017-02-08T00:00:00.000000000', 418005.414117),
                ('2017-03-15T00:00:00.000000000', 470489.149374),
            ],
            dtype=[('ds', '<M8[ns]'), ('y', '<f8')],
        )
    )
    with pytest.raises(ValueError):
        add_future_dates(test_data, periods=5)
        add_future_dates(test_data.iloc[:1, :], periods=5)


def test_add_future_dates():
    """Test that future dates are generated as expected."""
    test_data = pd.DataFrame.from_records(
        np.array(
            [
                ('2016-06-01T00:00:00.000000000', 453990.926731),
                ('2016-07-01T00:00:00.000000000', 461908.605069),
                ('2016-08-01T00:00:00.000000000', 482625.559603),
                ('2016-09-01T00:00:00.000000000', 438345.625953),
                ('2016-10-01T00:00:00.000000000', 458133.301185),
                ('2016-11-01T00:00:00.000000000', 467090.524589),
                ('2016-12-01T00:00:00.000000000', 508706.424433),
                ('2017-01-01T00:00:00.000000000', 426140.823832),
                ('2017-02-01T00:00:00.000000000', 418005.414117),
                ('2017-03-01T00:00:00.000000000', 470489.149374),
            ],
            dtype=[('ds', '<M8[ns]'), ('y', '<f8')],
        )
    )
    new_df = add_future_dates(test_data, periods=5)
    future_df = pd.DataFrame.from_records(
        np.array(
            [
                ('2017-04-01T00:00:00.000000000', np.nan),
                ('2017-05-01T00:00:00.000000000', np.nan),
                ('2017-06-01T00:00:00.000000000', np.nan),
                ('2017-07-01T00:00:00.000000000', np.nan),
                ('2017-08-01T00:00:00.000000000', np.nan),
            ],
            dtype=[('ds', '<M8[ns]'), ('y', '<f8')],
        )
    )
    expected_df = pd.concat([test_data, future_df]).reset_index(drop=True)
    pd.testing.assert_frame_equal(expected_df, new_df)


if __name__ == '__main__':
    unittest.main()



# cfg.py
"""
Configurations
----------
Configuration values for the SlackReport, MailReport, DBSaver and Mlflow.
"""
from typing import Optional

from decouple import AutoConfig
from muttlib.utils import get_default_jinja_template
from pkg_resources import resource_string

SQLITE = "sqlite"

SOAM_RUN_TABLE_BASENAME = "soam_flow_runs"
SOAM_RUN_FACTOR_CONF_TABLE_BASENAME = "soam_flow_run_factor_conf"
SOAM_TASK_RUNS_TABLE_BASENAME = "tasks_runs"
FORECASTER_VALUES_TABLE_BASENAME = "forecaster_values"
EXTRACT_VALUES_TABLE_BASENAME = "extract_values"

# Report paths and configs
MAIL_REPORT = "resources/mail_report.html"

# Template paths
TEMPLATE_DIR = "templates"
INIT_TEMPLATE = "init"
INIT_TEMPLATE_DIR = f"{TEMPLATE_DIR}/{INIT_TEMPLATE}"
TEMPLATES = [INIT_TEMPLATE]

# Text config
UTF_ENCODING = "utf-8"

# Table name setup
table_name_preffix = ""

SOAM_FLOW_RUN_TABLE = f"{table_name_preffix}{SOAM_RUN_TABLE_BASENAME}"
SOAM_FLOW_RUN_TABLE_FACTOR_CONF_TABLE = (
    f"{table_name_preffix}{SOAM_RUN_FACTOR_CONF_TABLE_BASENAME}"
)
SOAM_TASK_RUNS_TABLE = f"{table_name_preffix}{SOAM_TASK_RUNS_TABLE_BASENAME}"
FORECASTER_VALUES_TABLE = f"{table_name_preffix}{FORECASTER_VALUES_TABLE_BASENAME}"
# TODO: replace syntax to be length compliant
# FORECASTER_VALUES_TABLE = (f"{table_name_preffix}"
#                            f"{FORECASTER_VALUES_TABLE_BASENAME}")
EXTRACT_VALUES_TABLE = f"{table_name_preffix}{EXTRACT_VALUES_TABLE_BASENAME}"


# Mail report
MAIL_TEMPLATE = get_default_jinja_template(
    resource_string(__name__, MAIL_REPORT).decode(UTF_ENCODING)
).module

# Mlflow tracking config
# Set to True if tracking is on
TRACKING_IS_ACTIVE = False
# Local file paths should be prefixed with "file:/", default is ./mlruns
# http URIs and Databricks workspaces can be used too.
TRACKING_URI = ''


def get_db_cred(setting_path: Optional[str] = None) -> dict:
    """Read the setting.ini file and retrieve the database credentials

    Parameters
    ----------
    setting_path : str, optional
        The path for the .ini document with the settings.

    Returns
    -------
    dict
        A dict containing: username, dialect, password, database, port, host.
    """
    config = AutoConfig(search_path=setting_path)
    db_cred = {}

    db_cred["username"] = config("DB_USER")
    db_cred["dialect"] = config("DB_DIALECT")
    db_cred["password"] = config("DB_PASSWORD")
    db_cred["database"] = config("DB_NAME")
    db_cred["port"] = config("DB_PORT")
    db_cred["host"] = config("DB_IP")

    return db_cred


def get_slack_cred(setting_path: Optional[str] = None) -> dict:
    """Retrieve the Slack credentials

    Parameters
    ----------
    setting_path : str, optional
        The path for the .ini document with the settings.

    Returns
    -------
    dict
        A dict containing the slack token.
    """
    config = AutoConfig(search_path=setting_path)
    slack_creds = {}

    slack_creds["token"] = config("SLACK_TOKEN")

    return slack_creds


def get_smtp_cred(setting_path: Optional[str] = None) -> dict:
    """Retrieve the SMTP credentials

    Parameters
    ----------
    setting_path : str, optional
        The path for the .ini document with the settings.

    Returns
    -------
    dict
        A dict containing: user_address, password, mail_from, host, port.
    """
    config = AutoConfig(search_path=setting_path)
    smtp_creds = {}

    smtp_creds["user_address"] = config("SMTP_USER")
    smtp_creds["password"] = config("SMTP_PASS")
    smtp_creds["mail_from"] = config("SMTP_FROM")
    smtp_creds["host"] = config("SMTP_HOST")
    smtp_creds["port"] = config("SMTP_PORT")

    return smtp_creds


def get_db_uri(setting_path: Optional[str]) -> str:
    """Retrieve the uri for the database.

    Parameters
    ----------
    setting_path : str, optional
        The path for the .ini document with the settings.

    Returns
    -------
    str
        A string with the database uri.
    """
    db_cred = get_db_cred(setting_path)

    if db_cred["dialect"] == SQLITE:
        return f"{db_cred['dialect']}:///{db_cred['database']}"

    return f"{db_cred['dialect']}://{db_cred['username']}:{db_cred['password']}@{db_cred['host']}:{db_cred['port']}/{db_cred['database']}"
    # TODO: replace syntax to be length compliant
    # return (f"{db_cred['dialect']}://{db_cred['username']}:"
    #         f"db_cred['password']}@{db_cred['host']}:"
    #         f"{db_cred['port']}/{db_cred['database']}")



# savers.py
"""
Saver
-----
Saver is an abstract class used to store parameters and data through the
pipeline.
"""
from abc import ABC, abstractmethod
import logging

from prefect import Task
from prefect.engine.state import State

from soam.core import SoamFlow
from soam.data_models import StepTypeEnum
from soam.workflow import Forecaster

logger = logging.getLogger(__name__)


class Saver(ABC):  # pylint:disable=abstract-method
    """
    The base class for all savers objects.

    All implementations of Saver have to implement the state_handler of Prefect.
    Please check the [link](https://docs.prefect.io/core/concepts/states.html#state-handlers-callbacks)
    TODO: inherit from ABC or use ABCMeta to ensure abstractmethods implementation
    """

    @abstractmethod
    def __init__(self):  # pylint:disable=abstract-method
        pass

    @abstractmethod
    def save_forecast(self, task: Task, old_state: State, new_state: State):
        """
        This function will store the forecasts values.
        """

    @abstractmethod
    def save_task_run(self, task: Task, old_state: State, new_state: State):
        """
        This function will store the task run.
        """

    @abstractmethod
    def save_flow_run(self, soamflow: SoamFlow, old_state: State, new_state: State):
        """
        Will store the SoamFlow run.
        """

    def get_task_type(self, task: Task):
        """

        Parameters
        ----------
        task: Task
            Specified task to get the type of.

        Return
        ------
        str
            Step type of the task.
        """
        if isinstance(task, Forecaster):
            step_type = StepTypeEnum.forecast
        else:
            step_type = StepTypeEnum.custom
        return step_type



"""Wrapping functions for SkLearn API."""
import abc
from abc import abstractmethod
import inspect
from types import FunctionType
from typing import Callable, List

from sklearn.base import BaseEstimator


def get_clean_parameter_dict(signature: inspect.Signature) -> dict:
    """Remove self, kwarg and positional params from signature.

    Parameters
    ----------
    signature : inspect.Signature
        signature of a method

    Returns
    -------
    dict
        clean parameter dictionary
    """
    return {
        p.name: p.default if p.default != inspect.Parameter.empty else None
        for p in signature.parameters.values()
        if p.name != "self" and p.kind != p.VAR_KEYWORD and p.kind != p.VAR_POSITIONAL
    }


def sk_constructor_wrapper(modeltype) -> Callable:
    """Constructor patching decorator."""

    def sk_construct_function(init_func: Callable):
        """
        Join sklearn's and model's constructors.
        Ref: https://github.com/heidelbergcement/hcrystalball/blob/master/src/hcrystalball/wrappers/_base.py
        """
        orig_signature = inspect.signature(init_func)
        orig_parameters = get_clean_parameter_dict(orig_signature)

        model_signature = inspect.signature(modeltype.__init__)
        model_parameters = get_clean_parameter_dict(model_signature)

        full_parameter_names = list(model_parameters) + list(orig_parameters)
        full_parameter_defaults = list(model_parameters.values()) + list(
            orig_parameters.values()
        )
        assignments = "; ".join([f"self.{p}={p}" for p in full_parameter_names])

        constructor_code = compile(
            f'def __init__(self, {", ".join(full_parameter_names)}): ' f"{assignments}",
            "<string>",
            "exec",
        )

        modified_init_function = FunctionType(
            constructor_code.co_consts[0],  # code
            globals(),  # global dict
            "__init__",  # name of function
            tuple(full_parameter_defaults),  # defaults
        )

        return modified_init_function

    return sk_construct_function


class SkWrapper(BaseEstimator, metaclass=abc.ABCMeta):
    """Base class for model wrappers."""

    @abstractmethod
    def __init__(self):
        pass

    def _init_sk_model(
        self, model_class, clean=False, ignore_params: List[str] = None, **kwargs
    ) -> BaseEstimator:
        """Instantiate model_class.

        Parameters
        ----------
        model_class :
            Class of the model to wrap.
        clean : bool, optional
            Whether to check if param corresponds to
            the model's constructor or not, by default False
        ignore_params : List[str], optional
            Extra parameters specified in wrapper to be ignored, by default None
            i.e wrapper parameters that don't belong to the model.

        Returns
        -------
        Object's instance.
        """
        model_signature = inspect.signature(model_class.__init__)
        ignore_params = ignore_params or []
        model_params = orig_params = self.get_params()
        if clean:
            model_params = get_clean_parameter_dict(model_signature)
        params = {
            k: orig_params[k] for k, v in model_params.items() if k not in ignore_params
        }

        return model_class(**{**params, **kwargs})



# slack_report.py
"""
Slack Report
------------
Slack reporting and message formatting tools. Its a postprocess that sends the
model forecasts though the slack app.
"""
from asyncio import Future
from collections.abc import Iterable
from io import BytesIO
from pathlib import Path, PosixPath
from typing import IO, Any, Dict, Optional, Sequence, Union
import warnings

from jinja2 import Template
from muttlib.utils import path_or_string
import pandas as pd
import slack
from slack.web.slack_response import SlackResponse

from soam.cfg import get_slack_cred
from soam.constants import DS_COL, YHAT_COL
from soam.core.step import Step

DEFAULT_GREETING_MESSAGE = "Hello everyone! Here are the results of the forecast for the *{metric_name}* metric:\n"
DEFAULT_FAREWELL_MESSAGE = "Cheers!\n SoaM."


class SlackReport:
    """
    Generates the report to share via Slack.
    """

    warnings.warn(
        "This class will be deprecated in v1. Using SlackMessage and send_slack_message methods instead is recommended.",
        PendingDeprecationWarning,
    )

    def __init__(
        self, channel_id: str, metric_name: str, setting_path: Optional[str],
    ):
        """
        Initialization of the Slack Report object.

        Parameters
        ----------
        channel_id: str
            Slack channel id where the report will be sent.
        metric_name: str
            Performance metric being measured.
        setting_path: str
            Setting path.
        """
        credentials = get_slack_cred(setting_path)
        self.slack_client = slack.WebClient(credentials["token"])
        self.channel_id = channel_id
        self.metric_name = metric_name

    def send_report(
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ) -> Union[Future, SlackResponse]:
        """
        Send Slack report.

        Parameters
        ----------
        prediction : pd.DataFrame
            DataDrame of the predictions made.
        plot_filename : str or pathlib.Path
             Path of the forecast data to send.
        greeting_message: str
            Greeting message to send via Slack with the predictions.
        farewell_message: str
            Farewell message to send via Slack with the predictions.

        Returns
        -------
        Slack Report
            Sends the specified message with the predictions data via Slack.
        """
        if greeting_message == DEFAULT_GREETING_MESSAGE:
            greeting_message.format(metric_name=self.metric_name)

        prediction[f"{YHAT_COL}_str"] = prediction[YHAT_COL].apply(
            lambda f: "{:.2f}".format(f)  # pylint: disable=unnecessary-lambda
        )
        summary_message = _df_to_report_string(
            prediction,
            DS_COL,
            f"{YHAT_COL}_str",
            greeting_message=greeting_message,
            farewell_message=farewell_message,
        )
        prediction = prediction.drop(columns=[f"{YHAT_COL}_str"])

        return self.slack_client.files_upload(
            channels=self.channel_id,
            file=str(plot_filename),
            initial_comment=summary_message,
            title=f"{self.metric_name} Forecast",
        )


class SlackReportTask(Step, SlackReport):
    """
    Builds up the task of the report designed for Slack.
    """

    warnings.warn(
        "This class will be deprecated in v1. Using SlackMessage and send_slack_message methods instead is recommended.",
        PendingDeprecationWarning,
    )

    def __init__(
        self,
        channel_id: str,
        metric_name: str,
        setting_path: Optional[str],
        **kwargs: Any,
    ):
        """
        Parameters
        ----------
        channel_id: str
            Slack channel id where the report will be sent.
        metric_name: str
            Performance metric being measured.
        setting_path: str
            Setting path.
        kwargs:
            Extra args to pass.
        """
        Step.__init__(self, **kwargs)  # type: ignore
        SlackReport.__init__(
            self, channel_id, metric_name, setting_path,
        )

    def run(  # type: ignore
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ):
        """
        Send Slack report task.

        Parameters
        ----------
        prediction : pd.DataFrame
            DataDrame of the predictions made.
        plot_filename : str or pathlib.Path
             Path of the forecast data to send.
        greeting_message: str
            Greeting message to send via Slack with the predictions.
        farewell_message: str
            Farewell message to send via Slack with the predictions.

        Returns
        -------
        Slack Report Task
            Sends the specified message with the predictions data via Slack.
        """
        return self.send_report(
            prediction, plot_filename, greeting_message, farewell_message
        )


class SlackMessage:
    def __init__(
        self,
        template,
        arguments: Dict = None,
        attachment: Optional[Union[Path, IO]] = None,
    ):
        """Create Slack message.

        Args:
            template: Jinja template.
            template_args (dict): Arguments to be passed to the Jinja templates .
            attachment (Path): Path to file (or image) to attach to the image.
        """
        self.template = template
        self.arguments = arguments
        self.attachment_ref = attachment

    @property
    def message(self) -> str:
        """Message property."""
        message = path_or_string(self.template)
        if self.arguments:
            template = Template(message)
            message = template.render(**self.arguments)  # type:ignore
        return message

    @property
    def attachment(self) -> Optional[Union[str, IO]]:
        """Message property."""
        if self.attachment_ref is None:
            return None
        if isinstance(self.attachment_ref, PosixPath):
            if not self.attachment_ref.exists():
                raise ValueError(
                    f"File does not exist: {str(self.attachment_ref.resolve())}."
                )
            # slack's client supports a string with the path for the file
            return str(self.attachment_ref.resolve())
        elif isinstance(self.attachment_ref, BytesIO):
            return self.attachment_ref
        else:
            raise TypeError("Only PosixPath and BytesIO supported.")


def send_slack_message(
    slack_client: slack.WebClient,
    channel: str,
    msg: SlackMessage,
    thread_ts: Optional[int] = None,
):
    """Send Slack message.

    Parameters
    ----------
    channel : str
        slack channel to send the message to.
    msg : SlackMessage
        SlackMessage instance.
    thread_ts : int, optional
        message timestamp to reply to in threaded fashion, by default None.
    """
    if msg.attachment is None:
        response = slack_client.chat_postMessage(
            channel=channel, text=msg.message, thread_ts=thread_ts
        )
    else:
        response = slack_client.files_upload(
            file=msg.attachment,
            channels=channel,
            initial_comment=msg.message,
            thread_ts=thread_ts,
        )
    return response


def send_slack_messages_in_thread(
    slack_client: slack.WebClient, channel: str, messages: Sequence[SlackMessage]
):
    first_message = messages[0]
    response = send_slack_message(slack_client, channel, first_message)
    for msg in messages[1:]:
        response = send_slack_message(
            slack_client, channel, msg, thread_ts=response["ts"]
        )


def send_multiple_slack_messages(
    slack_client: slack.WebClient,
    channel: str,
    messages: Sequence[Union[SlackMessage, Sequence[SlackMessage]]],
):
    for element in messages:
        if isinstance(element, Iterable):
            send_slack_messages_in_thread(slack_client, channel, element)
        else:
            send_slack_message(slack_client, channel, element)


def _format_number(num):
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    # add more suffixes if you need them
    return '%.2f%s' % (num, ['', 'k', 'M', 'G', 'T', 'P'][magnitude])


def _df_to_report_string(
    df, date_col, value_col, greeting_message=None, farewell_message=None
):
    """Concatenates the rows using the date_col and value_col for formatting purposes."""
    summary_entries = []
    if greeting_message:
        summary_entries.append(greeting_message)

    for _, row in df.iterrows():
        date = row[date_col].strftime('%Y-%b-%d')
        value = row[value_col]
        summary_entries.append(f"• *[{date}]* {value}\n")

    if farewell_message:
        summary_entries.append(farewell_message)

    summary_message = "\n".join(summary_entries)  # type: ignore
    return summary_message


def send_anomaly_report(
    slack_client: slack.WebClient,
    channel_id: str,
    plot: Union[Path, IO],
    metric_name: str,
    anomaly_df: pd.DataFrame,
    date_col: str,
):
    """
    Parameters
    ----------
    channel_id: str
        Slack channel id where the report will be sent.
    plot: str, pathlib.Path or buffer
        Anomaly plot
    metric_name: str
        Metric to report.
    anomaly_df: pd.DataFrame
        DataFrame with anomalous values. Must have the following columns: ['y','yhat','yhat_lower','yhat_upper']
    date_col: str
        Name of the date column
    """
    detection_window = len(anomaly_df)
    stddev = anomaly_df.yhat.std()
    anomaly_df['stddev'] = abs(anomaly_df.yhat - anomaly_df.y) / stddev
    anomaly_df['anomaly_upper'] = anomaly_df['y'] > anomaly_df['yhat_upper']
    anomaly_df['anomaly_lower'] = anomaly_df['y'] < anomaly_df['yhat_lower']
    aux = anomaly_df.loc[(anomaly_df['anomaly_lower']) | (anomaly_df['anomaly_upper'])]
    if len(aux) == 0:
        greeting_message = "Hello everyone! Good news: there were no outliers found for the last {{detection_window}} days for the metric '{{metric_name}}'"
        msg = SlackMessage(
            greeting_message,
            arguments={
                "detection_window": detection_window,
                "metric_name": metric_name,
            },
            attachment=plot,
        )
    else:
        aux['message'] = aux.apply(
            lambda row: f"Count: {_format_number(row['y'])}, expected value in range ( {_format_number(row['yhat_lower'])} , {_format_number(row['yhat_upper'])} ) - {row['stddev']:.2f} standard deviations",
            axis=1,
        )
        greeting_message = f"Hello everyone! Here are the outliers found for the last {detection_window} days for the metric '{metric_name}'"
        summary_message = _df_to_report_string(
            aux,
            date_col,
            'message',
            greeting_message=greeting_message,
            farewell_message=None,
        )
        msg = SlackMessage(summary_message, attachment=plot)

    send_slack_message(slack_client, channel=channel_id, msg=msg)



# transformer.py
"""
Transformer
-----------
"""

import logging
from typing import TYPE_CHECKING, List, Optional, Tuple  # pylint:disable=unused-import

import pandas as pd
from prefect.utilities.tasks import defaults_from_attrs
from sklearn.base import BaseEstimator, TransformerMixin

from soam.core import Step

if TYPE_CHECKING:
    from soam.savers.savers import Saver


logger = logging.getLogger(__name__)


class BaseDataFrameTransformer(BaseEstimator, TransformerMixin):
    """
    Provide an interface to transform pandas DataFrames.
    """

    def fit(
        self, X: pd.DataFrame, **fit_params  # pylint:disable=unused-argument
    ) -> "BaseDataFrameTransformer":
        """
        Fit method

        Parameters
        ----------
        X: pandas.DataFrame
            DataFrame to fit.
        """
        logger.warning("Subclasses should implement this.")
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform method.

        Parameters
        ----------
            X: pd.DataFrame
                DataFrame to be transformed.

        Raises
        ------
            NotImplementedError
                Raises error if its not a subclass.

        Returns
        -------
            pd.DataFrame
        """
        raise NotImplementedError("Subclasses should implement this.")

    def fit_transform(self, X: pd.DataFrame, **fit_params) -> pd.DataFrame:
        """
        Fit Transform method.

        Parameters
        ----------
            X: pd.DataFrame
                DataFrame to be fitted and transformed.

        Returns
        -------
            pd.DataFrame
                DataFrame fitted and transformed.
        """
        return self.fit(X, **fit_params).transform(X)


class DummyDataFrameTransformer(BaseDataFrameTransformer):
    """
    This transformer provides an "identity transformation.
    Returns its input without any alteration.
    """

    def __init__(self):
        pass

    def fit(self, df_X):  # pylint:disable=unused-argument
        """
        Fit method

        Parameters
        ----------
            df_X: pandas.DataFrame
                DataFrame to be fitted.
        """
        return self

    def transform(self, df_X, inplace=True):
        """
        Transform method

        Parameters
        ----------
            df_X: pandas.DataFrame
                DataFrame to be transformed.
            inplace: bool
                Whether the changes should persist in the DataFrame or not. Default is True.

        Returns
        -------
            pandas.DataFrame
                DataFrame transformed.
        """
        if not inplace:
            df_X = df_X.copy()
        return df_X


class Transformer(Step):
    """
    Generates the transformer object.
    """

    transformer: BaseDataFrameTransformer

    def __init__(
        self,
        transformer: BaseDataFrameTransformer = None,
        savers: "Optional[List[Saver]]" = None,
        **kwargs
    ):
        """
        Handle transformations

        Parameters
        ----------
        transformer: BaseDataFrameTransformer
            Object that implements transformations a given dataset.
        savers : list of soam.savers.Saver, optional
            The saver to store the parameters and state changes.
        """
        super().__init__(**kwargs)
        if savers is not None:
            for saver in savers:
                self.state_handlers.append(saver.save_forecast)

        self.dataset = None
        self.transformed_dataset = None

        if transformer is None:
            transformer = DummyDataFrameTransformer()
        self.transformer = transformer

    def fit(self, dataset: pd.DataFrame) -> "Transformer":
        """
        Fit method

        Parameters
        ----------
            dataset: pandas.DataFrame
                DataFrame to be fitted.
        """
        self.transformer.fit(dataset)
        return self

    def transform(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Transform method

        Parameters
        ----------
            dataset: pandas.DataFrame
                DataFrame to be transformed.

        Returns
        -------
            pd.DataFrame
                DataFrame transformed.
        """
        return self.transformer.transform(dataset)

    def fit_transform(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """Fit Transform method.

        Parameters
        ----------
            dataset: pd.DataFrame
                DataFrame to be fitted and transformed.

        Returns
        -------
            pd.DataFrame
                DataFrame fitted and transformed.
        """
        return self.transformer.fit(dataset).transform(dataset)

    @defaults_from_attrs('transformer')
    def run(
        self, dataset: pd.DataFrame, transformer: BaseDataFrameTransformer = None
    ) -> Tuple[pd.DataFrame, Optional[BaseDataFrameTransformer]]:
        """
        Fit and apply a transformation on a dataset and return the fitted
        transformer and the new dataset.

        Parameters
        ----------
        dataset : pandas.DataFrame
            Dataset with data to be processed.

        Returns
        -------
        tuple of (pandas.DataFrame, BaseDataFrameTransformer)
            A tuple containing a pandas DataFrame with the transformed dataset
            and the fitted transformer.
        """
        self.dataset = dataset.copy()
        self.transformer = transformer  # type: ignore
        self.transformed_dataset = self.fit_transform(dataset)
        return self.transformed_dataset, self.transformer



"""Backtester"""
from copy import deepcopy
import unittest

import pandas as pd
import pytest
from sklearn.metrics import mean_absolute_error, mean_squared_error
from sklearn.preprocessing import StandardScaler

from soam.constants import (
    ANOMALY_PLOT,
    DS_COL,
    FIG_SIZE,
    MONTHLY_TIME_GRANULARITY,
    PLOT_CONFIG,
    Y_COL,
)
from soam.models.prophet import SkProphet
from soam.plotting.forecast_plotter import ForecastPlotterTask
from soam.workflow import (
    Backtester,
    BaseDataFrameTransformer,
    Forecaster,
    Transformer,
    compute_metrics,
)
from soam.workflow.backtester import METRICS_KEYWORD, PLOT_KEYWORD, RANGES_KEYWORD
from tests.helpers import sample_data_df  # pylint: disable=unused-import


def test_compute_metrics():
    """Function to compute performance metrics."""
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }
    y_true = [3, -0.5, 2, 7]
    y_pred = [2.5, 0.0, 2, 8]
    expected_output = {'mae': 0.5, 'mse': 0.375}
    output = compute_metrics(y_true, y_pred, metrics)
    unittest.TestCase().assertDictEqual(expected_output, output)


class SimpleProcessor(BaseDataFrameTransformer):
    """Create a Simple Processor object."""

    def __init__(self, **fit_params):  # pylint:disable=super-init-not-called
        self.preproc = StandardScaler(**fit_params)

    def fit(self, df_X):
        self.preproc.fit(df_X[Y_COL].values.reshape(-1, 1))
        return self

    def transform(self, df_X, inplace=True):
        if not inplace:
            df_X = df_X.copy()
        df_X[Y_COL] = self.preproc.transform(df_X[Y_COL].values.reshape(-1, 1)) + 10
        return df_X


def assert_backtest_fold_result_common_checks(rv, ranges=None, plots=None):
    """Backtest fold result common checks assertion."""
    assert tuple(rv) == (RANGES_KEYWORD, METRICS_KEYWORD, PLOT_KEYWORD)
    assert rv[RANGES_KEYWORD] == ranges
    assert rv[PLOT_KEYWORD].name == plots


def assert_backtest_fold_result(rv, ranges=None, metrics=None, plots=None):
    """Backtest fold result assertion."""
    assert_backtest_fold_result_common_checks(rv, ranges=ranges, plots=plots)
    for metric_name, values in metrics.items():
        assert metric_name in rv[METRICS_KEYWORD]
        if isinstance(values, dict):
            for measure_name, value in values.items():
                assert value, pytest.approx(rv[METRICS_KEYWORD][measure_name], 0.01)
        else:
            assert values, pytest.approx(rv[METRICS_KEYWORD][metric_name], 0.01)


def assert_backtest_all_folds_result(rvs, expected_values):
    """Backtest all fold result assertion."""
    assert len(rvs) == len(expected_values)
    for rv, evs in zip(rvs, expected_values):
        assert_backtest_fold_result(rv, **evs)


def assert_backtest_fold_result_aggregated(rv, ranges=None, metrics=None, plots=None):
    """Backtest fold result aggregated assertion."""
    assert_backtest_fold_result_common_checks(rv, ranges=ranges, plots=plots)
    output_metrics = pd.DataFrame(rv[METRICS_KEYWORD])
    expected_metrics = pd.DataFrame(metrics)
    pd.testing.assert_frame_equal(output_metrics, expected_metrics, rtol=1e-1)


def assert_backtest_all_folds_result_aggregated(rvs, expected_values):
    """Backtest all fold result aggregated assertion."""
    assert len(rvs) == len(expected_values)
    for rv, evs in zip(rvs, expected_values):
        assert_backtest_fold_result_aggregated(rv, **evs)


def test_integration_backtester_single_fold(
    tmp_path, sample_data_df
):  # pylint: disable=redefined-outer-name
    """Backtest single fold integration test."""
    test_window = 10
    train_data = sample_data_df
    forecaster = Forecaster(model=SkProphet(), output_length=test_window)
    preprocessor = Transformer(SimpleProcessor())
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=preprocessor,
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
    )
    rvs = backtester.run(train_data)
    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2015-07-01 00:00:00'),
                pd.Timestamp('2016-05-01 00:00:00'),
            ),
            METRICS_KEYWORD: {'mae': 0.19286372252777645, 'mse': 0.07077117049346579},
            'plots': '0_forecast_2013020100_2015080100_.png',
        },
    ]
    assert_backtest_all_folds_result(rvs, expected_values)


def test_integration_backtester_multi_fold(
    tmp_path, sample_data_df  # pylint: disable=redefined-outer-name
):
    """Backtest multi fold integration test."""
    test_window = 30
    train_data = pd.concat([sample_data_df] * 3)
    train_data[DS_COL] = pd.date_range(
        train_data[DS_COL].min(), periods=len(train_data), freq='MS'
    )
    model = SkProphet()
    forecaster = Forecaster(model=model, output_length=test_window)
    preprocessor = Transformer(SimpleProcessor())
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=preprocessor,
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
    )
    rvs = backtester.run(train_data)
    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2015-07-01 00:00:00'),
                pd.Timestamp('2018-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {'mae': 1.140921182444867, 'mse': 2.4605768804352675},
            'plots': '0_forecast_2013020100_2015080100_.png',
        },
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2015-08-01 00:00:00'),
                pd.Timestamp('2018-01-01 00:00:00'),
                pd.Timestamp('2020-07-01 00:00:00'),
            ),
            METRICS_KEYWORD: {'mae': 1.600049020613293, 'mse': 4.383723067139095},
            'plots': '0_forecast_2015080100_2018020100_.png',
        },
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2018-02-01 00:00:00'),
                pd.Timestamp('2020-07-01 00:00:00'),
                pd.Timestamp('2023-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {'mae': 3.1358162976127217, 'mse': 12.666965373730687},
            'plots': '0_forecast_2018020100_2020080100_.png',
        },
    ]
    assert_backtest_all_folds_result(rvs, expected_values)


# TODO: It maybe a good visual aggregation to include all metrics in one plot. This
# TODO: is not possible with the current implementation.
def test_integration_backtester_multi_fold_default_aggregation(
    tmp_path, sample_data_df  # pylint: disable=redefined-outer-name
):
    """Backtest multi fold default aggregation integration test."""
    test_window = 30
    train_data = pd.concat([sample_data_df] * 3)
    train_data[DS_COL] = pd.date_range(
        train_data[DS_COL].min(), periods=len(train_data), freq='MS'
    )
    model = SkProphet()
    forecaster = Forecaster(model=model, output_length=test_window)
    preprocessor = Transformer(SimpleProcessor())
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=preprocessor,
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
        aggregation="default",
    )
    rvs = backtester.run(train_data)
    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2023-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {
                'mae': {
                    'avg': 2.0269522786354313,
                    'max': 3.135813436023453,
                    'min': 1.344995687583762,
                },
                'mse': {
                    'avg': 6.761216280050696,
                    'max': 12.666927167728852,
                    'min': 3.233004063171241,
                },
            },
            'plots': '0_forecast_2018020100_2020080100_.png',
        }
    ]
    assert_backtest_all_folds_result_aggregated(rvs, expected_values)


def test_integration_backtester_multi_fold_custom_aggregations(
    tmp_path, sample_data_df  # pylint: disable=redefined-outer-name
):
    """Backtest multi fold custom aggregation integration test."""
    test_window = 30
    train_data = pd.concat([sample_data_df] * 3)
    train_data[DS_COL] = pd.date_range(
        train_data[DS_COL].min(), periods=len(train_data), freq='MS'
    )
    model = SkProphet()
    forecaster = Forecaster(model=model, output_length=test_window)
    preprocessor = Transformer(SimpleProcessor())
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }
    aggregation = {
        METRICS_KEYWORD: {
            "weighted_begining": lambda metrics_list: (
                sum(
                    [
                        3 * val if idx == 0 else val
                        for idx, val in enumerate(metrics_list)
                    ]
                )
                / (len(metrics_list) + 2)
            ),
            "weighted_ending": lambda metrics_list: (
                sum(
                    [
                        3 * val if idx == len(metrics_list) - 1 else val
                        for idx, val in enumerate(metrics_list)
                    ]
                )
                / (len(metrics_list) + 2)
            ),
        },
        PLOT_KEYWORD: 1,
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=preprocessor,
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
        aggregation=aggregation,
    )
    rvs = backtester.run(train_data)

    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2023-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {
                'mae': {
                    'weighted_begining': 1.631725773112123,
                    'weighted_ending': 2.4296838191792647,
                },
                'mse': {
                    'weighted_begining': 4.886483816435117,
                    'weighted_ending': 8.969039213753284,
                },
            },
            'plots': '0_forecast_2015080100_2018020100_.png',
        }
    ]
    assert_backtest_all_folds_result_aggregated(rvs, expected_values)


def test_integration_backtester_multi_fold_custom_metric_aggregation_default_plot(
    tmp_path, sample_data_df  # pylint: disable=redefined-outer-name
):
    """Backtest multi fold custom metric aggregation default plot integration test."""
    test_window = 30
    train_data = pd.concat([sample_data_df] * 3)
    train_data[DS_COL] = pd.date_range(
        train_data[DS_COL].min(), periods=len(train_data), freq='MS'
    )
    model = SkProphet()
    forecaster = Forecaster(model=model, output_length=test_window)
    preprocessor = Transformer(SimpleProcessor())
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }
    aggregation = {
        METRICS_KEYWORD: {
            "weighted_begining": lambda metrics_list: (
                sum(
                    [
                        3 * val if idx == 0 else val
                        for idx, val in enumerate(metrics_list)
                    ]
                )
                / (len(metrics_list) + 2)
            ),
            "weighted_ending": lambda metrics_list: (
                sum(
                    [
                        3 * val if idx == len(metrics_list) - 1 else val
                        for idx, val in enumerate(metrics_list)
                    ]
                )
                / (len(metrics_list) + 2)
            ),
        }
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=preprocessor,
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
        aggregation=aggregation,
    )
    rvs = backtester.run(train_data)

    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2023-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {
                'mae': {
                    'weighted_begining': 1.631725773112123,
                    'weighted_ending': 2.4296838191792647,
                },
                'mse': {
                    'weighted_begining': 4.886483816435117,
                    'weighted_ending': 8.969039213753284,
                },
            },
            'plots': '0_forecast_2018020100_2020080100_.png',
        }
    ]
    assert_backtest_all_folds_result_aggregated(rvs, expected_values)


def test_integration_backtester_multi_fold_custom_plot_aggregation_default_metric(
    tmp_path, sample_data_df  # pylint: disable=redefined-outer-name
):
    """Backtest multi fold default metric aggregation custom plot integration test."""
    test_window = 30
    train_data = pd.concat([sample_data_df] * 3)
    train_data[DS_COL] = pd.date_range(
        train_data[DS_COL].min(), periods=len(train_data), freq='MS'
    )
    model = SkProphet()
    forecaster = Forecaster(model=model, output_length=test_window)
    preprocessor = Transformer(SimpleProcessor())
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }
    aggregation = {
        PLOT_KEYWORD: 1,
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=preprocessor,
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
        aggregation=aggregation,
    )
    rvs = backtester.run(train_data)

    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2023-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {
                'mae': {
                    'avg': 2.0269522786354313,
                    'max': 3.135813436023453,
                    'min': 1.344995687583762,
                },
                'mse': {
                    'avg': 6.761216280050696,
                    'max': 12.666927167728852,
                    'min': 3.233004063171241,
                },
            },
            'plots': '0_forecast_2015080100_2018020100_.png',
        }
    ]
    assert_backtest_all_folds_result_aggregated(rvs, expected_values)
